package org.bitbucket.bradleysmithllc.etlunit.feature;

import com.google.inject.name.Named;
import org.bitbucket.bradleysmithllc.etlunit.*;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.parser.*;

import javax.inject.Inject;
import java.util.List;
import java.util.Map;

public class FeatureListenerProxy implements ClassListener
{
	private Log applicationLog;
	private boolean simulateOnly = false;

	private static final class ClassListenerFeature
	{
		final ClassListener listener;
		final Feature feature;

		private ClassListenerFeature(ClassListener listener, Feature feature)
		{
			this.listener = listener;
			this.feature = feature;
		}

		public Feature getFeature()
		{
			return feature;
		}

		public OperationProcessor getOperationProcessor()
		{
			return listener;
		}
	}

	private final FeatureProxyBase<ClassListenerFeature> base;

	@Inject
	public void setApplicationLog(@Named("applicationLog") Log log)
	{
		this.applicationLog = log;
	}

	@Inject
	public void setSimulateOnly(@Named("etlunit.simulateOnly") RuntimeOption ro)
	{
		this.simulateOnly = ro.isEnabled();
	}

	public FeatureListenerProxy(List<Feature> features)
	{
		base = new FeatureProxyBase<ClassListenerFeature>(features, new FeatureProxyBase.Getter<ClassListenerFeature>()
		{
			@Override
			public ClassListenerFeature get(Feature feature)
			{
				ClassListener classListener = feature.getListener();

				if (classListener != null)
				{
					return new ClassListenerFeature(classListener, feature);
				}

				return null;
			}
		});
	}

	private static interface FeatureListener
	{
		action_code broadcast(ClassListener director, Feature feature)
				throws TestAssertionFailure, TestExecutionError, TestWarning;
	}

	private action_code broadcast(FeatureListener dire) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		action_code ret = action_code.defer;

		for (ClassListenerFeature list : base.getList())
		{
			if (list != null)
			{
				if (!simulateOnly || list.feature.requiredForSimulation())
				{
					ret = dire.broadcast(list.listener, list.feature);

					if (ret == action_code.handled || ret == action_code.reject)
					{
						return ret;
					}
				}
			}
		}

		return ret;
	}

	@Override
	public void begin(final ETLTestClass cl, final VariableContext context)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		broadcast(new FeatureListener()
		{
			@Override
			public action_code broadcast(ClassListener director, Feature feature)
					throws TestAssertionFailure, TestExecutionError, TestWarning
			{
				director.begin(cl, context);

				return action_code.defer;
			}
		});
	}

	@Override
	public void declare(final ETLTestVariable var, final VariableContext context)
	{
		try
		{
			broadcast(new FeatureListener()
			{
				@Override
				public action_code broadcast(ClassListener director, Feature feature)
				{
					director.declare(var, context);

					return action_code.defer;
				}
			});

			// these are empty because the deferred method won't ever throw this
		}
		catch (TestAssertionFailure testAssertionFailure)
		{
		}
		catch (TestExecutionError testExecutionError)
		{
		}
		catch (TestWarning testWarning)
		{
		}
	}

	@Override
	public void begin(final ETLTestMethod mt, final VariableContext context)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		broadcast(new FeatureListener()
		{
			@Override
			public action_code broadcast(ClassListener director, Feature feature)
					throws TestAssertionFailure, TestExecutionError, TestWarning
			{
				director.begin(mt, context);

				return action_code.defer;
			}
		});
	}

	@Override
	public void begin(final ETLTestOperation op, final ETLTestValueObject parameters, final VariableContext vcontext, final ExecutionContext econtext)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		broadcast(new FeatureListener()
		{
			@Override
			public action_code broadcast(ClassListener director, Feature feature)
					throws TestAssertionFailure, TestExecutionError, TestWarning
			{
				director.begin(op, parameters, vcontext, econtext);

				return action_code.defer;
			}
		});
	}

	@Override
	public action_code process(ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		return processListenerList(base.getList(), op, obj, context, econtext, applicationLog, simulateOnly);
	}

	@Override
	public void end(final ETLTestOperation op, final ETLTestValueObject parameters, final VariableContext vcontext, final ExecutionContext econtext)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		broadcast(new FeatureListener()
		{
			@Override
			public action_code broadcast(ClassListener director, Feature feature)
					throws TestAssertionFailure, TestExecutionError, TestWarning
			{
				director.end(op, parameters, vcontext, econtext);

				return action_code.defer;
			}
		});
	}

	@Override
	public void end(final ETLTestMethod mt, final VariableContext context)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		broadcast(new FeatureListener()
		{
			@Override
			public action_code broadcast(ClassListener director, Feature feature)
					throws TestAssertionFailure, TestExecutionError, TestWarning
			{
				director.end(mt, context);

				return action_code.defer;
			}
		});
	}

	@Override
	public void end(final ETLTestClass cl, final VariableContext context)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		broadcast(new FeatureListener()
		{
			@Override
			public action_code broadcast(ClassListener director, Feature feature)
					throws TestAssertionFailure, TestExecutionError, TestWarning
			{
				director.end(cl, context);

				return action_code.defer;
			}
		});
	}

	public static action_code processListenerList(List<ClassListenerFeature> processorList, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext, Log applicationLog, boolean simulateOnly)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		action_code ret = action_code.defer;

		for (ClassListenerFeature processorObj : processorList)
		{
			if (!simulateOnly || processorObj.feature.requiredForSimulation())
			{
				ret = checkProcessor(processorObj, op, obj, context, econtext, applicationLog);

				if (ret == action_code.handled || ret == action_code.reject)
				{
					return ret;
				}
			}
		}

		if (ret == action_code.defer && simulateOnly)
		{
			ret = action_code.handled;
		}

		return ret;
	}

	private static action_code checkProcessor(ClassListenerFeature processor, ETLTestOperation op, ETLTestValueObject obj, VariableContext context, ExecutionContext econtext, Log applicationLog)
			throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		ETLTestValueObject obj_new = obj;

		ETLTestValueObject featureConfig = processor.getFeature().getFeatureConfiguration();

		if (featureConfig != null)
		{
			// merge it into the supplied parameters, preferring the operation parameters
			if (obj_new != null)
			{
				obj_new = obj_new.merge(featureConfig, ETLTestValueObject.merge_type.left_merge);
			}
			else
			{
				obj_new = featureConfig;
			}
		}

		// always call the lowest-level process handler - which is optional
		OperationProcessor operationProcessor = processor.getOperationProcessor();
		if (operationProcessor != null)
		{
			action_code res = operationProcessor.process(op, obj, context, econtext);

			if (res == action_code.reject || res == action_code.handled)
			{
				return res;
			}
		}

		// check if validation is possible
		FeatureMetaInfo meta = processor.getFeature().getMetaInfo();

		// this is a change from 1.2 - utilize the meta-data to determine if a feature can use an operation
		Map<String, FeatureOperation> operations = meta.getExportedOperations();

		if (operations != null)
		{
			FeatureOperation featureOperation = operations.get(op.getOperationName());

			if (featureOperation != null)
			{
				return featureOperation.process(op, obj_new, context, econtext);
			}
		}

		return action_code.defer;
	}
}