package org.bitbucket.bradleysmithllc.etlunit.feature;

import com.fasterxml.jackson.databind.JsonNode;
import com.github.fge.jackson.JsonLoader;
import com.github.fge.jsonschema.exceptions.ProcessingException;
import com.github.fge.jsonschema.main.JsonSchema;
import com.github.fge.jsonschema.report.ProcessingReport;
import org.bitbucket.bradleysmithllc.etlunit.ExecutionContext;
import org.bitbucket.bradleysmithllc.etlunit.TestAssertionFailure;
import org.bitbucket.bradleysmithllc.etlunit.TestExecutionError;
import org.bitbucket.bradleysmithllc.etlunit.TestWarning;
import org.bitbucket.bradleysmithllc.etlunit.context.VariableContext;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestOperation;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

public class ResourceFeatureOperation implements FeatureOperation
{
	private final String name;
	private final String description;
	private final ResourceFeatureMetaInfo resourceFeatureMetaInfo;
	private final List<String> signatures;
	private final List<FeatureOperationSignature> operationSignatures = new ArrayList<FeatureOperationSignature>();

	public static final JsonNode NULL_NODE;

	static
	{
		try
		{
			NULL_NODE = JsonLoader.fromString("{}");
		}
		catch (IOException e)
		{
			throw new RuntimeException(e);
		}
	}

	public ResourceFeatureOperation(ResourceFeatureMetaInfo resourceFeatureMetaInfo, String key, String description, List<String> signatur)
	{
		name = key;
		this.description = description;
		this.resourceFeatureMetaInfo = resourceFeatureMetaInfo;
		this.signatures = signatur;

		for (String sig : signatures)
		{
			operationSignatures.add(new ResourceFeatureOperationSignature(this, resourceFeatureMetaInfo, sig));
		}
	}

	@Override
	public String getName()
	{
		return name;
	}

	@Override
	public String getDescription()
	{
		return description;
	}

	@Override
	public String getUsage()
	{
		String val = resourceFeatureMetaInfo.getResource(name + ".usage");

		return val;
	}

	@Override
	public String getPrototype()
	{
		String val = resourceFeatureMetaInfo.getResource(name + ".prototype");

		return val;
	}

	@Override
	public List<FeatureOperationSignature> getSignatures()
	{
		return operationSignatures;
	}

	@Override
	public action_code process(ETLTestOperation op, ETLTestValueObject parameters, VariableContext vcontext, ExecutionContext econtext) throws TestAssertionFailure, TestExecutionError, TestWarning
	{
		if (name.equals(op.getOperationName()))
		{
			for (FeatureOperationSignature sig : operationSignatures)
			{
				JsonSchema schema = sig.getValidator();

				if (schema != null)
				{
					resourceFeatureMetaInfo.getApplicationLog().info("Verifying operation request '" + op.getOperationName() + "' against the schema signature '" + sig.getId() + "'");

					try
					{
						ProcessingReport report = schema.validate(parameters != null ? parameters.getJsonNode() : NULL_NODE);

						if (report.isSuccess())
						{
							return sig.process(op, parameters, vcontext, econtext);
						}
					}
					catch (ProcessingException e)
					{
						// didn't validate - pass on to the next one
					}
				}
				else
				{
					// no schema, this one wins
					return sig.process(op, parameters, vcontext, econtext);
				}
			}
		}

		return action_code.defer;
	}
}